import base64


def b64url_decode(data: str) -> bytes:
    padding = "=" * (-len(data) % 4)
    return base64.urlsafe_b64decode(data + padding)


def b64url_encode(data: bytes) -> str:
    """Base64url encode bytes."""
    return base64.urlsafe_b64encode(data).rstrip(b"=").decode()



import pytest
from pydantic import BaseModel

from src.usso_jwt import algorithms, exceptions, schemas, sign


@pytest.fixture
def test_key() -> algorithms.AbstractKey:
    return algorithms.EdDSAKey.generate()


@pytest.fixture
def test_token(
    test_valid_payload: dict,
    test_header: dict,
    test_key: algorithms.AbstractKey,
):
    jwt = sign.generate_jwt(
        header=test_header,
        payload=test_valid_payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    return jwt


def test_jwt(
    test_token: str,
    test_key: algorithms.AbstractKey,
    test_header: dict,
    test_valid_payload: dict,
):
    jwt_obj = schemas.JWT(
        token=test_token,
        config=schemas.JWTConfig(key=test_key.jwk()),
    )
    assert jwt_obj.header == test_header
    assert jwt_obj.payload == test_valid_payload
    assert jwt_obj.verify()


def test_invalid_token(test_token: str, test_key: algorithms.AbstractKey):
    with pytest.raises(exceptions.JWTInvalidFormatError):
        invalid_token = f"{test_token[:-2]}.{test_token[-1]}"
        jwt_obj = schemas.JWT(
            token=invalid_token,
            config=schemas.JWTConfig(key=test_key.jwk()),
        )
        jwt_obj.verify()


def test_payload_class(
    test_token: str, test_key: algorithms.AbstractKey, test_valid_payload: dict
):
    class TestPayload(BaseModel):
        sub: str
        name: str
        iat: int
        exp: int

    jwt_obj = schemas.JWT(
        token=test_token,
        config=schemas.JWTConfig(key=test_key.jwk()),
        payload_class=TestPayload,
    )

    print(jwt_obj._parts[1])
    assert jwt_obj.payload == TestPayload(
        **test_valid_payload,
    )


def test_not_verified_payload(
    test_token: str, test_key: algorithms.AbstractKey
):
    jwt_obj = schemas.JWT(
        token=test_token[:-1],
        config=schemas.JWTConfig(key=test_key.jwk()),
    )
    with pytest.raises(exceptions.JWTInvalidFormatError):
        print(jwt_obj.token, jwt_obj.payload)


def test_no_key(test_token: str):
    with pytest.raises(ValueError):
        jwt_obj = schemas.JWT(
            token=test_token,
        )
        jwt_obj.verify()


def test_is_temporally_valid_true(
    test_token: str, test_key: algorithms.AbstractKey
):
    jwt_obj = schemas.JWT(
        token=test_token,
        config=schemas.JWTConfig(key=test_key.jwk()),
    )
    assert jwt_obj.is_temporally_valid() is True


def test_is_temporally_valid_false_missing_claims(
    test_header: dict, test_key: algorithms.AbstractKey
):
    # Payload missing temporal claims
    payload = {"sub": "user1", "name": "Test User"}
    token = sign.generate_jwt(
        header=test_header,
        payload=payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    jwt_obj = schemas.JWT(
        token=token,
        config=schemas.JWTConfig(key=test_key.jwk()),
    )
    assert jwt_obj.is_temporally_valid()


def test_is_temporally_valid_false_invalid_claims(
    test_header: dict, test_key: algorithms.AbstractKey
):
    # Payload with expired exp claim
    import time

    payload = {
        "sub": "user1",
        "name": "Test User",
        "exp": int(time.time()) - 100,
    }
    token = sign.generate_jwt(
        header=test_header,
        payload=payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    jwt_obj = schemas.JWT(
        token=token,
        config=schemas.JWTConfig(key=test_key.jwk()),
    )
    assert jwt_obj.is_temporally_valid() is False



from .base import AbstractKey, KeyAlgorithm
from .ecdsa import ECDSAAlgorithm, ECDSAKey
from .eddsa import EdDSAAlgorithm, EdDSAKey
from .hmac import HMACAlgorithm, HMACKey
from .rsa import RSAAlgorithm, RSAKey


def get_algorithm(alg: str) -> type[KeyAlgorithm]:
    """
    Get the appropriate algorithm manager for the given algorithm.

    Args:
        alg: The algorithm name
             (e.g., "HS256", "RS256", "ES256", "EdDSA", "Ed25519")

    Returns:
        The appropriate algorithm manager class

    Raises:
        ValueError: If the algorithm is not supported
    """
    for algo in KeyAlgorithm.__subclasses__():
        if alg.upper() in algo.SUPPORTED_ALGORITHMS:
            return algo

    raise ValueError(f"Unsupported algorithm: {alg}")


__all__ = [
    "KeyAlgorithm",
    "AbstractKey",
    "HMACKey",
    "RSAKey",
    "ECDSAKey",
    "EdDSAKey",
    "HMACAlgorithm",
    "RSAAlgorithm",
    "ECDSAAlgorithm",
    "EdDSAAlgorithm",
    "get_algorithm",
]



import json_advanced as json
from cryptography.hazmat.primitives.asymmetric import ec, ed25519, rsa

from .algorithms import get_algorithm
from .utils import b64url_encode


def create_jwt_header(alg: str, kid: str | None = None, **kwargs) -> dict:
    header = {"alg": alg, "typ": "JWT"}
    if kid:
        header["kid"] = kid
    header.update(kwargs)
    return header


def sign_jwt_parts(
    *,
    key: (
        dict
        | bytes
        | rsa.RSAPrivateKey
        | ec.EllipticCurvePrivateKey
        | ed25519.Ed25519PrivateKey
    ),
    alg: str,
    signing_input: str | bytes | None = None,
    header: dict | None = None,
    payload: dict | None = None,
    password: bytes | None = None,
) -> str:
    """Sign JWT parts using the specified algorithm."""
    # Prepare signing input
    if header is None and payload is None and signing_input is None:
        raise ValueError(
            "Either header, payload, or signing_input must be provided",
        )
    if signing_input is None:
        header_b64 = b64url_encode(json.dumps(header).encode())
        payload_b64 = b64url_encode(json.dumps(payload).encode())
        signing_input = f"{header_b64}.{payload_b64}".encode()

    # Get algorithm and sign
    algorithm = get_algorithm(alg)
    signature = algorithm.sign(
        data=signing_input,
        key=key,
        alg=alg,
        password=password,
    )
    return signature


def generate_jwt(
    header: dict,
    payload: dict,
    key: dict | bytes,
    alg: str,
    password: bytes | None = None,
) -> str:
    header_b64 = b64url_encode(json.dumps(header).encode())
    payload_b64 = b64url_encode(json.dumps(payload).encode())
    signing_input = f"{header_b64}.{payload_b64}"
    signature = sign_jwt_parts(
        key=key,
        alg=alg,
        signing_input=signing_input.encode(),
        password=password,
    )

    # Return complete JWT
    signature_b64 = b64url_encode(signature)
    return f"{signing_input}.{signature_b64}"



import json_advanced as json
import pytest

from src.usso_jwt import sign, verify
from src.usso_jwt.algorithms import AbstractKey
from src.usso_jwt.utils import b64url_encode


@pytest.fixture
def test_header(test_key: AbstractKey) -> dict:
    return sign.create_jwt_header(alg=test_key.algorithm, kid=test_key.kid)


def test_private_pem(
    test_key: AbstractKey, test_header: dict, test_valid_payload: dict
):
    pem = test_key.private_pem(password=b"123456")
    key = AbstractKey.load_pem(pem, password=b"123456")
    assert pem
    signature = sign.sign_jwt_parts(
        header=test_header,
        payload=test_valid_payload,
        key=key,
        alg=test_key.algorithm,
    )
    assert signature is not None

    header_b64 = b64url_encode(json.dumps(test_header).encode())
    payload_b64 = b64url_encode(json.dumps(test_valid_payload).encode())
    signing_input = f"{header_b64}.{payload_b64}".encode()

    assert verify.verify_signature(
        alg=test_key.algorithm,
        key=test_key.public_der(),
        signature=signature,
        data=signing_input,
    )


def test_sign_verify(
    test_valid_payload: dict, test_header: dict, test_key: AbstractKey
):
    signature = sign.sign_jwt_parts(
        header=test_header,
        payload=test_valid_payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    # print(signature, test_key.algorithm)
    # print(test_key.public_pem().decode())
    header_b64 = b64url_encode(json.dumps(test_header).encode())
    payload_b64 = b64url_encode(json.dumps(test_valid_payload).encode())
    signing_input = f"{header_b64}.{payload_b64}".encode()

    assert signature is not None
    assert verify.verify_signature(
        alg=test_key.algorithm,
        key=test_key.public_der(),
        signature=signature,
        data=signing_input,
    )
    assert verify.verify_temporal_claims(payload=test_valid_payload)


def test_generate_jwt(
    test_valid_payload: dict, test_header: dict, test_key: AbstractKey
):
    jwt = sign.generate_jwt(
        header=test_header,
        payload=test_valid_payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    assert jwt is not None
    header, payload, signature, signing_input = verify.extract_jwt_parts(jwt)
    assert header == test_header
    assert payload == test_valid_payload
    assert signature is not None
    assert verify.verify_signature(
        alg=test_key.algorithm,
        key=test_key.public_der(),
        signature=signature,
        data=signing_input,
    )
    # assert verify.verify_temporal_claims(payload=payload)
    assert verify.verify_jwt(token=jwt, jwk=test_key.jwk())



from enum import StrEnum


class Algorithm(StrEnum):
    # HMAC with SHA-2
    HS256 = "HS256"  # HMAC with SHA-256
    HS384 = "HS384"  # HMAC with SHA-384
    HS512 = "HS512"  # HMAC with SHA-512

    # RSA with SHA-2
    RS256 = "RS256"  # RSA with SHA-256
    RS384 = "RS384"  # RSA with SHA-384
    RS512 = "RS512"  # RSA with SHA-512

    # ECDSA with SHA-2
    ES256 = "ES256"  # ECDSA with SHA-256
    ES384 = "ES384"  # ECDSA with SHA-384
    ES512 = "ES512"  # ECDSA with SHA-512

    # RSASSA-PSS with SHA-2
    PS256 = "PS256"  # RSASSA-PSS with SHA-256
    PS384 = "PS384"  # RSASSA-PSS with SHA-384
    PS512 = "PS512"  # RSASSA-PSS with SHA-512

    # EdDSA with SHA-2
    EdDSA = "EDDSA"  # EdDSA with SHA-512
    Ed25519 = "ED25519"  # EdDSA with SHA-512

    @property
    def kty(self) -> str:
        return {
            Algorithm.HS256: "oct",
            Algorithm.HS384: "oct",
            Algorithm.HS512: "oct",
            Algorithm.RS256: "RSA",
            Algorithm.RS384: "RSA",
            Algorithm.RS512: "RSA",
            Algorithm.ES256: "EC",
            Algorithm.ES384: "EC",
            Algorithm.ES512: "EC",
            Algorithm.PS256: "RSA",
            Algorithm.PS384: "RSA",
            Algorithm.PS512: "RSA",
            Algorithm.EdDSA: "OKP",
            Algorithm.Ed25519: "OKP",
        }[self]



import json
import os
from typing import Any

from pydantic import BaseModel, Field, field_validator, model_validator

from .enums import Algorithm


class JWTConfig(BaseModel):
    """Configuration for JWT processing."""

    jwk_url: str | None = Field(
        default=None,
        description="URL to fetch JWK from. Required if key is not provided.",
    )
    key: str | bytes | dict | None = Field(
        default=None,
        description="""Key for signing/verification. Can be a PEM string,
        bytes, or JWK dict. If PEM string, it will be converted to JWK.
        If JWK dict, it will be used as is.
        If bytes, it will be converted to JWK.
        """,
    )
    issuer: str | list[str] | None = Field(
        default=None,
        description="Expected issuer(s) of the JWT.",
    )
    audience: str | list[str] | None = Field(
        default=None,
        description="Expected audience(s) of the JWT.",
    )
    type: Algorithm = Field(
        default=Algorithm.EdDSA,
        description="Algorithm to use for signing/verification.",
    )

    def __init__(self, **data: Any):
        if os.getenv("JWT_CONFIG") and not data:
            data = json.loads(os.getenv("JWT_CONFIG", "{}"))

        super().__init__(**data)

    @classmethod
    def init_by_json(cls, json_data: str | dict) -> "JWTConfig":
        if isinstance(json_data, str):
            json_data = json.loads(json_data)
        return cls(**json_data)

    def __hash__(self) -> int:
        return hash(self.model_dump_json())

    @model_validator(mode="after")
    def validate_config(cls, data: "JWTConfig") -> "JWTConfig":
        if not data.jwk_url and not data.key:
            raise ValueError("Either jwk_url or key must be provided")
        return data

    @field_validator("key", mode="after")
    def validate_key(cls, v: dict | str | bytes | None) -> dict | None:
        from .algorithms.base import convert_key_to_jwk

        if v is None:
            return None
        if isinstance(v, str):
            return convert_key_to_jwk(v.encode())
        if isinstance(v, bytes):
            return convert_key_to_jwk(v)
        return v



import os

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, hmac

from ..utils import b64url_decode, b64url_encode
from .base import AbstractKey, KeyAlgorithm


class HMACAlgorithm(KeyAlgorithm):
    """HMAC algorithm implementation (HS256, HS384, HS512)."""

    SUPPORTED_ALGORITHMS = {
        "HS256": hashes.SHA256,
        "HS384": hashes.SHA384,
        "HS512": hashes.SHA512,
    }

    @staticmethod
    def load_key(key: dict | bytes, password: bytes | None = None) -> bytes:
        """
        Load HMAC key from JWK dict or raw bytes.

        Args:
            key: Either a JWK dict or raw key bytes
            password: Optional password for encrypted keys

        Returns:
            HMAC key bytes
        """
        if isinstance(key, dict):
            return b64url_decode(key["k"])
        return key

    @classmethod
    def sign(
        cls,
        *,
        data: bytes,
        key: dict | bytes,
        alg: str = "HS256",
        password: bytes | None = None,
    ) -> bytes:
        """
        Sign using HMAC algorithms.

        Args:
            data: The data to sign
            key: Either a JWK dict or raw key bytes
            alg: The signing algorithm to use (HS256, HS384, HS512)
            password: Optional password for encrypted keys

        Returns:
            The signature
        """
        if alg not in cls.SUPPORTED_ALGORITHMS:
            raise ValueError(f"Unsupported HMAC algorithm: {alg}")

        key_bytes = cls.load_key(key, password)
        h = hmac.HMAC(
            key_bytes,
            cls.SUPPORTED_ALGORITHMS[alg](),
            backend=default_backend(),
        )
        h.update(data)
        return h.finalize()

    @classmethod
    def verify(
        cls,
        *,
        data: bytes,
        signature: bytes,
        key: dict | bytes,
        alg: str = "HS256",
        password: bytes | None = None,
    ) -> bool:
        """
        Verify HMAC signature.

        Args:
            data: The data that was signed
            signature: The signature to verify
            key: Either a JWK dict or raw key bytes
            alg: The signing algorithm used (HS256, HS384, HS512)
            password: Optional password for encrypted keys

        Returns:
            True if signature is valid, False otherwise
        """
        if alg not in cls.SUPPORTED_ALGORITHMS:
            raise ValueError(f"Unsupported HMAC algorithm: {alg}")

        key_bytes = cls.load_key(key, password)
        h = hmac.HMAC(
            key_bytes,
            cls.SUPPORTED_ALGORITHMS[alg](),
            backend=default_backend(),
        )
        h.update(data)
        try:
            h.verify(signature)
            return True
        except Exception:
            return False


class HMACKey(AbstractKey):
    """HMAC key implementation."""

    SUPPORTED_ALGORITHMS = {
        "HS256": hashes.SHA256,
        "HS384": hashes.SHA384,
        "HS512": hashes.SHA512,
    }

    def __init__(self, *, key: bytes, algorithm: str = "HS256"):
        self.key = key
        self.algorithm = algorithm

    @classmethod
    def generate(
        cls,
        *,
        algorithm: str = "HS256",
        key_size: int = 32,
    ) -> "HMACKey":
        """Generate a new HMAC key."""
        return cls(
            key=os.urandom(key_size),
            algorithm=algorithm,
        )

    @classmethod
    def load_jwk(cls, key: dict) -> "HMACKey":
        """Load a key from JWK dict."""
        algorithm = key.get("alg", "HS256")
        return cls(
            key=b64url_decode(key["k"]),
            algorithm=algorithm,
        )

    @classmethod
    def load_pem(
        cls,
        key: bytes,
        password: bytes | None = None,
        algorithm: str = "HS256",
    ) -> "HMACKey":
        """Load a key from PEM."""
        key = super().load_pem(key, password)
        return cls(key=key, algorithm=algorithm)

    @classmethod
    def load_der(
        cls,
        key: bytes,
        password: bytes | None = None,
        algorithm: str = "HS256",
    ) -> "HMACKey":
        """Load a key from DER."""
        key = super().load_der(key, password)
        return cls(key=key, algorithm=algorithm)

    def public_key(self) -> bytes:
        """Get the public key."""
        return self.key

    @property
    def jwk(self, kid: str | None = None) -> dict:
        """Get the JWK for the key."""
        return {  # type: ignore
            "kty": "oct",
            "alg": self.algorithm,
            "k": b64url_encode(self.key),
            "kid": kid or self.kid,
        }

    @property
    def type(self) -> str:
        """Get the type of the key."""
        return "HMAC"

    @property
    def key_size(self) -> int:
        """Get the size of the key."""
        return len(self.key)

    def sign(self, data: bytes) -> bytes:
        """Sign data using the key."""
        return HMACAlgorithm.sign(data=data, key=self.key, alg=self.algorithm)

    def verify(self, data: bytes, signature: bytes) -> bool:
        """Verify signature using the key."""
        return HMACAlgorithm.verify(
            data=data, signature=signature, key=self.key, alg=self.algorithm
        )



"""JWT-related exceptions."""


class JWTError(Exception):
    """Base exception for JWT-related errors."""

    def __init__(self, message: str = None, *args):
        self.message = message
        super().__init__(message, *args)


class JWTExpiredError(JWTError):
    """Raised when a JWT has expired."""

    def __init__(self):
        self.message = "JWT has expired"
        super().__init__(self.message)


class JWTNotValidYetError(JWTError):
    """Raised when a JWT's 'nbf' claim indicates it's not valid yet."""

    def __init__(self):
        self.message = "JWT is not valid yet"
        super().__init__(self.message)


class JWTIssuedInFutureError(JWTError):
    """Raised when a JWT's 'iat' claim is in the future."""

    def __init__(self):
        self.message = "JWT is issued in the future"
        super().__init__(self.message)


class JWTInvalidSignatureError(JWTError):
    """Raised when a JWT's signature is invalid."""

    def __init__(self):
        self.message = "JWT signature is invalid"
        super().__init__(self.message)


class JWTInvalidFormatError(JWTError):
    """Raised when a JWT has an invalid format."""

    def __init__(self):
        self.message = "JWT has an invalid format"
        super().__init__(self.message)


class JWKNotFoundError(JWTError):
    """Raised when a JWK with the specified kid is not found."""

    def __init__(self):
        self.message = "JWK with the specified kid is not found"
        super().__init__(self.message)


class JWTInvalidAudienceError(JWTError):
    """Raised when a JWT's audience claim is invalid."""

    def __init__(self):
        self.message = "JWT audience claim is invalid"
        super().__init__(self.message)


class JWTInvalidTokenTypeError(JWTError):
    """Raised when a JWT's token type claim is invalid."""

    def __init__(
        self,
        expected_token_type: str | None = None,
        provided_token_type: str | None = None,
    ):
        self.message = "JWT token type claim is invalid"
        if expected_token_type and provided_token_type:
            self.message += (
                f" (expected: {expected_token_type},"
                f" provided: {provided_token_type})"
            )
        elif expected_token_type:
            self.message += f" (expected: {expected_token_type})"
        elif provided_token_type:
            self.message += f" (provided: {provided_token_type})"
        super().__init__(self.message)


class JWTInvalidACRError(JWTError):
    """Raised when a JWT's acr claim is invalid."""

    def __init__(self):
        self.message = "JWT acr claim is invalid"
        super().__init__(self.message)


class JWTInvalidIssuerError(JWTError):
    """Raised when a JWT's issuer claim is invalid."""

    def __init__(self):
        self.message = "JWT issuer claim is invalid"
        super().__init__(self.message)


class JWTMissingAudienceError(JWTError):
    """Raised when a JWT's audience claim is required but missing."""

    def __init__(self):
        self.message = "JWT audience claim is required but missing"
        super().__init__(self.message)
